@@ -6,12 +6,14 @@
 import io
 import logging
 import os
+import shutil
 import tempfile
 import zipfile
 from pathlib import Path
-from typing import Any, List, Optional
+from typing import Any, List, Optional, cast
 
 import py7zr
+import rarfile
 from natsort import natsorted, ns
 from PIL import Image
 
@@ -22,6 +24,70 @@ from .genericmetadata import GenericMetadata
 logger = logging.getLogger(__name__)
 
 
+class UnknownArchiver:
+
+    """Unknown implementation"""
+
+    def __init__(self, path: Path) -> None:
+        self.path = path
+
+    def read_file(self, archive_file: str) -> Any:
+        return Any
+
+    def write_file(self, archive_file: str, data: str) -> bool:
+        return False
+
+    def remove_file(self, archive_file: str) -> bool:
+        return False
+
+    def get_filename_list(self) -> List[str]:
+        return []
+
+    def copy_from_archive(self, other_archive: "UnknownArchiver") -> bool:
+        return False
+
+
+# ------------------------------------------------------------------
+class RarArchiver:
+    """Rar implementation"""
+
+    def __init__(self, path: Path) -> None:
+        self.path = path
+
+    def read_file(self, archive_file: str) -> bytes:
+        """Read the contents of a comic archive"""
+        archive = rarfile.RarFile(self.path)
+        try:
+            return archive.read(archive_file)
+        except rarfile.RarCannotExec as e:
+            logger.error(f"Error reading rar archive [{e}]: {self.path} :: {archive_file}")
+            raise rarfile.RarCannotExec from e
+
+    def remove_file(self) -> bool:
+        """Rar files are read-only, so we return False."""
+        return False
+
+    def write_file(self) -> bool:
+        """Rar files are read-only, so we return False."""
+        return False
+
+    def get_filename_list(self) -> List[str]:
+        """Returns a list of the filenames in an archive"""
+        archive = rarfile.RarFile(self.path)
+        try:
+            return sorted(archive.namelist())
+        except rarfile.RarCannotExec as e:
+            logger.error(f"Error reading rar archive [{e}]: {self.path}")
+            raise rarfile.RarCannotExec from e
+
+    def copy_from_archive(self) -> bool:
+        """Rar files are read-only, so we return False."""
+        return False
+
+
+# ------------------------------------------------------------------
+
+
 class ZipArchiver:
 
     """ZIP implementation"""
@@ -29,88 +95,88 @@ class ZipArchiver:
     def __init__(self, path: Path) -> None:
         self.path = path
 
-    def read_archive_file(self, archive_file: str) -> bytes:
+    def read_file(self, archive_file: str) -> bytes:
         """Read the contents of a comic archive"""
-
-        data = ""
-        zip_file = zipfile.ZipFile(self.path, "r")
-
         try:
-            data = zip_file.read(archive_file)
-        except zipfile.BadZipfile as bad_zip_error:
-            logger.exception(f"bad zipfile [{bad_zip_error}]: {self.path} :: {archive_file}")
-            zip_file.close()
-            raise OSError from bad_zip_error
-        except Exception as exception_error:
-            logger.exception(f"bad zipfile [{exception_error}]: {self.path} :: {archive_file}")
-            zip_file.close()
-            raise OSError from exception_error
-        finally:
-            zip_file.close()
-        return data
+            data = bytes()
+            with zipfile.ZipFile(self.path, mode="r") as zf:
+                data = zf.read(archive_file)
+            return data
+        except (zipfile.BadZipfile, OSError) as e:
+            logger.error(f"Error reading zip archive [{e}]: {self.path} :: {archive_file}")
+            raise OSError from e
 
-    def remove_archive_file(self, archive_file: str) -> bool:
+    def remove_file(self, archive_file: str) -> bool:
         """Returns a boolean when attempting to remove a file from an archive"""
+        return self._rebuild_zipfile([archive_file])
 
-        try:
-            self._rebuild_zipfile([archive_file])
-        except zipfile.BadZipfile:
-            return False
-        else:
-            return True
-
-    def write_archive_file(self, archive_file: str, data: str) -> bool:
+    def write_file(self, archive_file: str, data: str) -> bool:
         #  At the moment, no other option but to rebuild the whole
         #  zip archive w/o the indicated file. Very sucky, but maybe
         # another solution can be found
-        try:
-            self._rebuild_zipfile([archive_file])
+        files = self.get_filename_list()
+        if archive_file in files:
+            self.rebuild([archive_file])
 
+        try:
             # now just add the archive file as a new one
-            zip_file = zipfile.ZipFile(
+            with zipfile.ZipFile(
                 self.path, mode="a", allowZip64=True, compression=zipfile.ZIP_DEFLATED
-            )
-            zip_file.writestr(archive_file, data)
-            zip_file.close()
+            ) as zf:
+                zf.writestr(archive_file, data)
             return True
-        except (zipfile.BadZipfile, zipfile.LargeZipFile) as exception_error:
-            logger.exception(f"Error writing zipfile: {exception_error}.")
+        except (zipfile.BadZipfile, OSError) as e:
+            logger.error(f"Error writing zip archive [{e}]: {self.path} :: {archive_file}")
             return False
 
-    def get_archive_filename_list(self) -> List[str]:
+    def get_filename_list(self) -> List[str]:
         """Returns a list of the filenames in an archive"""
-
         try:
-            zip_file = zipfile.ZipFile(self.path, "r")
-            namelist = zip_file.namelist()
-            zip_file.close()
+            with zipfile.ZipFile(self.path, mode="r") as zf:
+                namelist = zf.namelist()
             return namelist
-        except Exception as exception_error:
-            logger.exception(f"Unable to get zipfile list [{exception_error}]: {self.path}")
+        except (zipfile.BadZipfile, OSError) as e:
+            logger.error(f"Error listing files in zip archive [{e}]: {self.path}")
             return []
 
     def _rebuild_zipfile(self, exclude_list: List[str]) -> None:
-        """Zip helper func
+        """
+        Zip helper func
 
         This recompresses the zip archive, without the files in the exclude_list
         """
-        # generate temp file
-        tmp_fd, tmp_name = tempfile.mkstemp(dir=self.path.parent)
-        os.close(tmp_fd)
+        try:
+            with zipfile.ZipFile(
+                tempfile.NamedTemporaryFile(dir=self.path.parent, delete=False),
+                "w",
+                allowZip64=True,
+            ) as zout:
+                with zipfile.ZipFile(self.path, mode="r") as zin:
+                    for item in zin.infolist():
+                        buffer = zin.read(item.filename)
+                        if item.filename not in exclude_list:
+                            zout.writestr(item, buffer)
 
-        zin = zipfile.ZipFile(self.path, "r")
-        zout = zipfile.ZipFile(tmp_name, "w", allowZip64=True)
-        for item in zin.infolist():
-            buffer = zin.read(item.filename)
-            if item.filename not in exclude_list:
-                zout.writestr(item, buffer)
+                # replace with the new file
+                self.path.unlink()
+                shutil.move(cast(str, zout.filename), self.path)
+            return True
+        except (zipfile.BadZipfile, OSError) as e:
+            logger.error(f"Error rebuilding zip file [{e}]: {self.path}")
+            return False
 
-        zout.close()
-        zin.close()
-
-        # replace with the new file
-        self.path.unlink()
-        os.rename(tmp_name, self.path)
+    def copy_from_archive(self, other_archive: UnknownArchiver) -> bool:
+        """Replace the current zip with one copied from another archive"""
+        try:
+            with zipfile.ZipFile(self.path, mode="w", allowZip64=True) as zout:
+                for filename in other_archive.get_filename_list():
+                    data = other_archive.read_file(filename)
+                    if data is not None:
+                        zout.writestr(filename, data)
+            return True
+        except Exception as e:
+            logger.warning(f"Error while copying to {self.path}: {e}")
+            return False
 
 
 # ------------------------------------------------------------------
@@ -121,115 +187,76 @@ class SevenZipArchiver:
     def __init__(self, path: Path) -> None:
         self.path = path
 
-    def read_archive_file(self, archive_file: str):
-        data = ""
+    def read_file(self, archive_file: str) -> bytes:
+        """Read the contents of a comic archive"""
         try:
+            data = bytes()
             with py7zr.SevenZipFile(self.path, "r") as zf:
-                if r := zf.read([archive_file]):
-                    data = r[archive_file].read()
-                else:
-                    logger.warning(f"No data in 7zipfile: {self.path} :: {archive_file}")
-                    raise OSError
-        except py7zr.Bad7zFile as e:
-            logger.warning(f"bad 7zip file [{e}]: {self.path} :: {archive_file}")
-            raise OSError from e
-        except Exception as e:
+                data = zf.read(archive_file)[archive_file].read()
+            return data
+        except (py7zr.Bad7zFile, OSError) as e:
             logger.warning(f"bad 7zip file [{e}]: {self.path} :: {archive_file}")
             raise OSError from e
 
-        return data
+    def remove_file(self, archive_file: str) -> bool:
+        """Returns a boolean when attempting to remove a file from an archive"""
+        return self._rebuild([archive_file])
+
+    def write_file(self, archive_file: str, data: str) -> bool:
+        files = self.get_filename_list()
+        if archive_file in files:
+            self._rebuild([archive_file])
 
-    def remove_archive_file(self, archive_file: str) -> bool:
         try:
-            self._rebuild_seven_zipfile([archive_file])
-        except py7zr.Bad7zFile:
-            return False
-        else:
-            return True
-
-    def write_archive_file(self, archive_file: str, data: str) -> bool:
-        try:
-            files = self.get_archive_filename_list()
-            if archive_file in files:
-                self._rebuild_seven_zipfile([archive_file])
-
             # now just add the archive file as a new one
             with py7zr.SevenZipFile(self.path, "a") as zf:
                 zf.writestr(data, archive_file)
             return True
-        except py7zr.Bad7zFile:
+        except (py7zr.Bad7zFile, OSError) as e:
+            logger.error(f"Error writing 7zip archive [{e}]: {self.path} :: {archive_file}")
             return False
 
-    def get_archive_filename_list(self) -> List[str]:
+    def get_filename_list(self) -> List[str]:
+        """Returns a list of the filenames in an archive"""
         try:
             with py7zr.SevenZipFile(self.path, "r") as zf:
                 namelist = zf.getnames()
-
             return namelist
-        except Exception as e:
-            logger.warning("Unable to get 7zip file list [%s]: %s", e, self.path)
+        except (py7zr.Bad7zFile, OSError) as e:
+            logger.warning(f"Unable to get 7zip file list [{e}]: {self.path}")
             return []
 
-    def _rebuild_seven_zipfile(self, exclude_list: List[str]) -> None:
+    def _rebuild(self, exclude_list: List[str]) -> bool:
         """Zip helper func
 
         This recompresses the zip archive, without the files in the exclude_list
         """
-        tmp_fd, tmp_name = tempfile.mkstemp(dir=self.path.parent)
-        os.close(tmp_fd)
-
         try:
             with py7zr.SevenZipFile(self.path, "r") as zip:
                 targets = [f for f in zip.getnames() if f not in exclude_list]
-            with py7zr.SevenZipFile(self.path, "r") as zin:
-                with py7zr.SevenZipFile(tmp_name, "w") as zout:
-                    if r := zin.read(targets):
-                        for fname, bio in r.items():
-                            zout.writef(bio, fname)
+            with tempfile.NamedTemporaryFile(dir=self.path.parent, delete=False) as tmp_file:
+                with py7zr.SevenZipFile(tmp_file.file, mode="w") as zout:
+                    with py7zr.SevenZipFile(self.path, mode="r") as zin:
+                        for filename, buffer in zin.read(targets).items():
+                            zout.writef(buffer, filename)
+                self.path.unlink()
+                shutil.move(tmp_file.name, self.path)
+            return True
         except Exception as e:
             logger.warning("Exception[%s]: %s", e, self.path)
-            return
+            return False
 
-        # replace with the new file
-        self.path.unlink()
-        os.rename(tmp_name, self.path)
-
-    def copy_from_archive(self, other_archive) -> bool:
+    def copy_from_archive(self, other_archive: UnknownArchiver) -> bool:
         """Replace the current zip with one copied from another archive"""
         try:
             with py7zr.SevenZipFile(self.path, "w") as zout:
-                for fname in other_archive.get_archive_filename_list():
-                    if data := other_archive.read_archive_file(fname):
+                for fname in other_archive.get_filename_list():
+                    if data := other_archive.read_file(fname):
                         zout.writestr(data, fname)
-        except Exception as e:
-            logger.warning("Error while copying to %s: %s", self.path, e)
-            return False
-        else:
             return True
-
-
-# ------------------------------------------------------------------
-class UnknownArchiver:
-
-    """Unknown implementation"""
-
-    def __init__(self, path: Path) -> None:
-        self.path = path
-
-    def read_archive_file(self, archive_file: str) -> Any:
-        return Any
-
-    def write_archive_file(self, archive_file: str, data: str) -> bool:
-        return False
-
-    def remove_archive_file(self, archive_file: str) -> bool:
-        return False
-
-    def get_archive_filename_list(self) -> List[str]:
-        return []
-
-    def copy_from_archive(self) -> bool:
-        return False
+        except Exception as e:
+            logger.warning(f"Error while copying to {self.path}: {e}")
+            return False
 
 
 # ------------------------------------------------------------------
@@ -242,7 +269,7 @@ class ComicArchive:
     class ArchiveType:
         """Types of archives supported. Currently only .cbz and .cb7"""
 
-        zip, sevenzip, unknown = list(range(3))
+        zip, sevenzip, rar, unknown = list(range(4))
 
     def __init__(self, path: Path) -> None:
         self.path = path
@@ -259,6 +286,9 @@ class ComicArchive:
         elif self.zip_test():
             self.archive_type: int = self.ArchiveType.zip
             self.archiver = ZipArchiver(self.path)
+        elif self.rar_test():
+            self.archive_type: int = self.ArchiveType.rar
+            self.archiver = RarArchiver(self.path)
         else:
             self.archive_type = self.ArchiveType.unknown
             self.archiver = UnknownArchiver(self.path)
@@ -270,6 +300,10 @@ class ComicArchive:
         self.page_list = None
         self.metadata = None
 
+    def rar_test(self) -> bool:
+        """Test whether an archive is a rar file"""
+        return rarfile.is_rarfile(self.path)
+
     def sevenzip_test(self) -> bool:
         """Tests whether an archive is a sevenzipfile"""
         return py7zr.is_7zfile(self.path)
@@ -279,6 +313,10 @@ class ComicArchive:
 
         return zipfile.is_zipfile(self.path)
 
+    def is_rar(self) -> bool:
+        """Returns a boolean as to whether an archive is a rarfile."""
+        return self.archive_type == self.ArchiveType.rar
+
     def is_sevenzip(self) -> bool:
         """Returns a boolean as to whether an archive is a sevenzipfile"""
         return self.archive_type == self.ArchiveType.sevenzip
@@ -294,12 +332,18 @@ class ComicArchive:
         if self.archive_type == self.ArchiveType.unknown:
             return False
 
+        if self.is_rar():
+            return False
+
         return bool(os.access(self.path, os.W_OK))
 
     def seems_to_be_a_comic_archive(self) -> bool:
         """Returns a boolean as to whether the file is a comic archive"""
 
-        return bool((self.is_zip() or self.is_sevenzip()) and (self.get_number_of_pages() > 0))
+        return bool(
+            (self.is_zip() or self.is_sevenzip() or self.is_rar())
+            and (self.get_number_of_pages() > 0)
+        )
 
     def get_page(self, index: int) -> Optional[bytes]:
         """Returns an image(page) from an archive"""
@@ -310,7 +354,7 @@ class ComicArchive:
 
         if filename is not None:
             try:
-                image_data = self.archiver.read_archive_file(filename)
+                image_data = self.archiver.read_file(filename)
             except OSError:
                 logger.exception("Error reading in page.")
 
@@ -339,7 +383,7 @@ class ComicArchive:
 
         if self.page_list is None:
             # get the list file names in the archive, and sort
-            files = self.archiver.get_archive_filename_list()
+            files = self.archiver.get_filename_list()
 
             # seems like some archive creators are on  Windows, and don't know
             # about case-sensitivity!
@@ -387,7 +431,7 @@ class ComicArchive:
         if not self.has_metadata():
             return None
         try:
-            tmp_raw_metadata = self.archiver.read_archive_file(self.ci_xml_filename)
+            tmp_raw_metadata = self.archiver.read_file(self.ci_xml_filename)
             # Convert bytes to str. Is it safe to decode with utf-8?
             raw_metadata = tmp_raw_metadata.decode("utf-8")
         except OSError:
@@ -397,19 +441,21 @@ class ComicArchive:
 
     def write_metadata(self, metadata: Optional[GenericMetadata]) -> bool:
         """Write the metadata to the archive"""
-
         if metadata is None:
             return False
         self.apply_archive_info_to_metadata(metadata, calc_page_sizes=True)
-        md_string = ComicInfoXml().string_from_metadata(metadata)
-        write_success = self.archiver.write_archive_file(self.ci_xml_filename, md_string)
+        if raw_cix := self.read_raw_metadata():
+            md_string = ComicInfoXml().string_from_metadata(metadata, raw_cix.encode("utf-8"))
+        else:
+            md_string = ComicInfoXml().string_from_metadata(metadata)
+        write_success = self.archiver.write_file(self.ci_xml_filename, md_string)
         return self._successful_write(write_success, True, metadata)
 
     def remove_metadata(self) -> bool:
         """Remove the metadata from the archive if present"""
 
         if self.has_metadata():
-            write_success = self.archiver.remove_archive_file(self.ci_xml_filename)
+            write_success = self.archiver.remove_file(self.ci_xml_filename)
             return self._successful_write(write_success, False, None)
         return True
 
@@ -430,7 +476,7 @@ class ComicArchive:
                 self.seems_to_be_a_comic_archive()
                 and (
                     not self.seems_to_be_a_comic_archive()
-                    or self.ci_xml_filename in self.archiver.get_archive_filename_list()
+                    or self.ci_xml_filename in self.archiver.get_filename_list()
                 )
             )
 
@@ -489,9 +535,20 @@ class ComicArchive:
         return metadata
 
     def export_as_cb7(self, new_7zip_filename: Path) -> bool:
+        """Export CBZ archive to CB7 format."""
+        if self.archive_type == self.ArchiveType.sevenzip:
+            # nothing to do, we're already a 7zip
+            return True
+
+        zip_archiver = SevenZipArchiver(new_7zip_filename)
+
+        return zip_archiver.copy_from_archive(self.archiver)
+
+    def export_as_zip(self, zipfilename: Path) -> bool:
         if self.archive_type == self.ArchiveType.zip:
-            zip_archiver = SevenZipArchiver(new_7zip_filename)
-        else:
-            return False
+            # nothing to do, we're already a zip
+            return True
+
+        zip_archiver = ZipArchiver(zipfilename)
 
         return zip_archiver.copy_from_archive(self.archiver)
